@@ -1,3 +1,4 @@
+# coding: utf8
 """
     cssselect.xpath
     ===============
@@ -82,6 +83,14 @@ class XPathExpr(object):
 
 split_at_single_quotes = re.compile("('+)").split
 
+# The spec is actually more permissive than that, but don’t bother.
+# This is just for the fast path.
+# http://www.w3.org/TR/REC-xml/#NT-NameStartChar
+is_safe_name = re.compile('^[a-zA-Z_][a-zA-Z0-9_.-]*$').match
+
+# Test that the string is not empty and does not contain whitespace
+is_non_whitespace = re.compile(r'^[^ \t\r\n\f]+$').match
+
 
 #### Translation
 
@@ -115,6 +124,10 @@ class GenericTranslator(object):
     #: http://www.w3.org/TR/selectors/#id-selectors
     id_attribute = 'id'
 
+    #: The attribute used for ``:lang()`` depends on the document language:
+    #: http://www.w3.org/TR/selectors/#lang-pseudo
+    lang_attribute = 'xml:lang'
+
     #: The case sensitivity of document language element names,
     #: attribute names, and attribute values in selectors depends
     #: on the document language.
@@ -176,7 +189,9 @@ class GenericTranslator(object):
         tree = getattr(selector, 'parsed_tree', None)
         if not tree:
             raise TypeError('Expected a parsed selector, got %r' % (selector,))
-        return (prefix or '') + _unicode(self.xpath(tree))
+        xpath = self.xpath(tree)
+        assert isinstance(xpath, XPathExpr)  # help debug a missing 'return'
+        return (prefix or '') + _unicode(xpath)
 
     @staticmethod
     def xpath_literal(s):
@@ -245,15 +260,19 @@ class GenericTranslator(object):
             name = selector.attrib.lower()
         else:
             name = selector.attrib
-        if selector.namespace == '*':
-            name = '@' + name
+        safe = is_safe_name(name)
+        if selector.namespace:
+            name = '%s:%s' % (selector.namespace, name)
+            safe = safe and is_safe_name(selector.namespace)
+        if safe:
+            attrib = '@' + name
         else:
-            name = '@%s:%s' % (selector.namespace, name)
+            attrib = 'attribute::*[name() = %s]' % self.xpath_literal(name)
         if self.lower_case_attribute_values:
             value = selector.value.lower()
         else:
             value = selector.value
-        return method(self.xpath(selector.selector), name, value)
+        return method(self.xpath(selector.selector), attrib, value)
 
     def xpath_class(self, class_selector):
         """Translate a class selector."""
@@ -269,15 +288,23 @@ class GenericTranslator(object):
 
     def xpath_element(self, selector):
         """Translate a type or universal selector."""
-        if self.lower_case_element_names:
-            element = selector.element.lower()
+        element = selector.element
+        if not element:
+            element = '*'
+            safe = True
         else:
-            element = selector.element
-        if selector.namespace != '*':
+            safe = is_safe_name(element)
+            if self.lower_case_element_names:
+                element = element.lower()
+        if selector.namespace:
             # Namespace prefixes are case-sensitive.
             # http://www.w3.org/TR/css3-namespace/#prefixes
             element = '%s:%s' % (selector.namespace, element)
-        return XPathExpr(element=element)
+            safe = safe and is_safe_name(selector.namespace)
+        xpath = XPathExpr(element=element)
+        if not safe:
+            xpath.add_name_test()
+        return xpath
 
 
     # CombinedSelector: dispatch by combinator
@@ -362,14 +389,29 @@ class GenericTranslator(object):
                                              add_name_test=False)
 
     def xpath_contains_function(self, xpath, function):
-        return xpath.add_condition('contains(string(.), %s)'
-                            % self.xpath_literal(function.arguments))
+        # Defined there, removed in later drafts:
+        # http://www.w3.org/TR/2001/CR-css3-selectors-20011113/#content-selectors
+        if function.argument_types() not in (['STRING'], ['IDENT']):
+            raise ExpressionError(
+                "Expected a single string or ident for :contains(), got %r"
+                % function.arguments)
+        value = function.arguments[0].value
+        return xpath.add_condition(
+            'contains(string(.), %s)' % self.xpath_literal(value))
 
-    def function_unsupported(self, xpath, pseudo):
-        raise ExpressionError(
-            "The pseudo-class :%s() is not supported" % pseudo.name)
-
-    xpath_lang_function = function_unsupported
+    def xpath_lang_function(self, xpath, function):
+        if function.argument_types() not in (['STRING'], ['IDENT']):
+            raise ExpressionError(
+                "Expected a single string or ident for :lang(), got %r"
+                % function.arguments)
+        value = function.arguments[0].value
+        return xpath.add_condition(
+            "ancestor-or-self::*[@lang][1][starts-with(concat("
+                # XPath 1.0 has no lower-case function...
+                "translate(@%s, 'ABCDEFGHIJKLMNOPQRSTUVWXYZ', "
+                               "'abcdefghijklmnopqrstuvwxyz'), "
+                "'-'), %s)]"
+            % (self.lang_attribute, self.xpath_literal(value.lower() + '-')))
 
 
     # Pseudo: dispatch by pseudo-class name
@@ -451,9 +493,12 @@ class GenericTranslator(object):
         return xpath
 
     def xpath_attrib_includes(self, xpath, name, value):
-        xpath.add_condition(
-            "%s and contains(concat(' ', normalize-space(%s), ' '), %s)"
-            % (name, name, self.xpath_literal(' '+value+' ')))
+        if is_non_whitespace(value):
+            xpath.add_condition(
+                "%s and contains(concat(' ', normalize-space(%s), ' '), %s)"
+                % (name, name, self.xpath_literal(' '+value+' ')))
+        else:
+            xpath.add_condition('0')
         return xpath
 
     def xpath_attrib_dashmatch(self, xpath, name, value):
@@ -465,19 +510,31 @@ class GenericTranslator(object):
         return xpath
 
     def xpath_attrib_prefixmatch(self, xpath, name, value):
-        return xpath.add_condition('%s and starts-with(%s, %s)' % (
-            name, name, self.xpath_literal(value)))
+        if value:
+            xpath.add_condition('%s and starts-with(%s, %s)' % (
+                name, name, self.xpath_literal(value)))
+        else:
+            xpath.add_condition('0')
+        return xpath
 
     def xpath_attrib_suffixmatch(self, xpath, name, value):
-        # Oddly there is a starts-with in XPath 1.0, but not ends-with
-        return xpath.add_condition(
-            '%s and substring(%s, string-length(%s)-%s) = %s'
-            % (name, name, name, len(value)-1, self.xpath_literal(value)))
+        if value:
+            # Oddly there is a starts-with in XPath 1.0, but not ends-with
+            xpath.add_condition(
+                '%s and substring(%s, string-length(%s)-%s) = %s'
+                % (name, name, name, len(value)-1, self.xpath_literal(value)))
+        else:
+            xpath.add_condition('0')
+        return xpath
 
     def xpath_attrib_substringmatch(self, xpath, name, value):
-        # Attribute selectors are case sensitive
-        return xpath.add_condition('%s and contains(%s, %s)' % (
-            name, name, self.xpath_literal(value)))
+        if value:
+            # Attribute selectors are case sensitive
+            xpath.add_condition('%s and contains(%s, %s)' % (
+                name, name, self.xpath_literal(value)))
+        else:
+            xpath.add_condition('0')
+        return xpath
 
 
 class HTMLTranslator(GenericTranslator):
@@ -496,6 +553,9 @@ class HTMLTranslator(GenericTranslator):
         are case-insensitive.
 
     """
+
+    lang_attribute = 'lang'
+
     def __init__(self, xhtml=False):
         self.xhtml = xhtml  # Might be useful for sub-classes?
         if not xhtml:
